@@ -1,6 +1,6 @@
 -- |
 -- Module      :  Path.IO
--- Copyright   :  © 2016–2019 Mark Karpov
+-- Copyright   :  © 2016–present Mark Karpov
 -- License     :  BSD 3 clause
 --
 -- Maintainer  :  Mark Karpov <markkarpov92@gmail.com>
@@ -12,7 +12,6 @@
 -- scanning and copying of directories, working with temporary
 -- files\/directories, etc.
 
-{-# LANGUAGE CPP               #-}
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE TemplateHaskell   #-}
 {-# LANGUAGE TupleSections     #-}
@@ -47,10 +46,10 @@ module Path.IO
   , getAppUserDataDir
   , getUserDocsDir
   , getTempDir
-#if MIN_VERSION_directory(1,2,3)
-  , XdgDirectory (..)
+  , D.XdgDirectory (..)
   , getXdgDir
-#endif
+  , D.XdgDirectoryList (..)
+  , getXdgDirList
     -- * Path transformation
   , AnyPath (..)
   , resolveFile
@@ -66,6 +65,10 @@ module Path.IO
   , findFiles
   , findFilesWith
     -- * Symbolic links
+  , createFileLink
+  , createDirLink
+  , removeDirLink
+  , getSymlinkTarget
   , isSymlink
     -- * Temporary files and directories
   , withTempFile
@@ -96,11 +99,9 @@ module Path.IO
   , setPermissions
   , copyPermissions
     -- * Timestamps
-#if MIN_VERSION_directory(1,2,3)
   , getAccessTime
   , setAccessTime
   , setModificationTime
-#endif
   , getModificationTime )
 where
 
@@ -112,6 +113,7 @@ import Control.Monad.Trans.Class (lift)
 import Control.Monad.Trans.Maybe (MaybeT (..), runMaybeT)
 import Control.Monad.Trans.Writer.Lazy (WriterT, execWriterT, tell)
 import Data.Either (lefts, rights)
+import Data.Kind (Type)
 import Data.List ((\\))
 import Data.Time (UTCTime)
 import Path
@@ -124,10 +126,6 @@ import qualified System.FilePath          as F
 import qualified System.IO.Temp           as T
 import qualified System.PosixCompat.Files as P
 
-#if MIN_VERSION_directory(1,2,3)
-import System.Directory (XdgDirectory)
-#endif
-
 ----------------------------------------------------------------------------
 -- Actions on directories
 
@@ -386,6 +384,15 @@ listDirRecurRel dir = (DList.toList *** DList.toList)
 -- ones.
 --
 -- __Note__: before version /1.3.0/, this function followed symlinks.
+--
+-- __Note__: before version /1.6.0/, the function created empty directories
+-- in the destination directory when the source directory contained
+-- directory symlinks. The symlinked directories were not recursively
+-- traversed. It also copied symlinked files creating normal regular files
+-- in the target directory as the result. This was fixed in the version
+-- /1.6.0/ so that the function now behaves much like the @cp@ utility, not
+-- traversing symlinked directories, but recreating symlinks in the target
+-- directory according to their targets in the source directory.
 
 copyDirRecur :: (MonadIO m, MonadCatch m)
   => Path b0 Dir       -- ^ Source
@@ -400,6 +407,15 @@ copyDirRecur = copyDirRecurGen True
 -- @since 1.1.0
 --
 -- __Note__: before version /1.3.0/, this function followed symlinks.
+--
+-- __Note__: before version /1.6.0/, the function created empty directories
+-- in the destination directory when the source directory contained
+-- directory symlinks. The symlinked directories were not recursively
+-- traversed. It also copied symlinked files creating normal regular files
+-- in the target directory as the result. This was fixed in the version
+-- /1.6.0/ so that the function now behaves much like the @cp@ utility, not
+-- traversing symlinked directories, but recreating symlinks in the target
+-- directory according to their targets in the source directory.
 
 copyDirRecur' :: (MonadIO m, MonadCatch m)
   => Path b0 Dir       -- ^ Source
@@ -409,16 +425,14 @@ copyDirRecur' = copyDirRecurGen False
 
 -- | Generic version of 'copyDirRecur'. The first argument controls whether
 -- to preserve directory permissions or not. /Does not/ follow symbolic
--- links.
---
--- __Note__: before version /1.3.0/, this function followed symlinks.
+-- links. Internal function.
 
 copyDirRecurGen :: MonadIO m
   => Bool              -- ^ Should we preserve directory permissions?
   -> Path b0 Dir       -- ^ Source
   -> Path b1 Dir       -- ^ Destination
   -> m ()
-copyDirRecurGen p src dest = liftIO $ do
+copyDirRecurGen preserveDirPermissions src dest = liftIO $ do
   bsrc  <- makeAbsolute src
   bdest <- makeAbsolute dest
   (dirs, files) <- listDirRecur bsrc
@@ -429,14 +443,28 @@ copyDirRecurGen p src dest = liftIO $ do
         -> IO (Path Abs t)
       swapParent old new path = (new </>) <$>
         stripProperPrefix old path
-  tdirs  <- mapM (swapParent bsrc bdest) dirs
-  tfiles <- mapM (swapParent bsrc bdest) files
   ensureDir bdest
-  mapM_ ensureDir tdirs
-  zipWithM_ copyFile files tfiles
-  when p $ do
+  forM_ dirs $ \srcDir -> do
+    destDir <- swapParent bsrc bdest srcDir
+    dirIsSymlink <- isSymlink srcDir
+    if dirIsSymlink
+      then do
+        target <- getSymlinkTarget srcDir
+        D.createDirectoryLink target $
+          F.dropTrailingPathSeparator (toFilePath destDir)
+      else ensureDir destDir
+    when preserveDirPermissions $
+      ignoringIOErrors (copyPermissions srcDir destDir)
+  forM_ files $ \srcFile -> do
+    destFile <- swapParent bsrc bdest srcFile
+    fileIsSymlink <- isSymlink srcFile
+    if fileIsSymlink
+      then do
+        target <- getSymlinkTarget srcFile
+        D.createFileLink target (toFilePath destFile)
+      else copyFile srcFile destFile
+  when preserveDirPermissions $
     ignoringIOErrors (copyPermissions bsrc bdest)
-    zipWithM_ (\s d -> ignoringIOErrors $ copyPermissions s d) dirs tdirs
 
 ----------------------------------------------------------------------------
 -- Walking directory trees
@@ -836,15 +864,14 @@ getUserDocsDir = liftIO $ D.getUserDocumentsDirectory >>= parseAbsDir
 getTempDir :: MonadIO m => m (Path Abs Dir)
 getTempDir = liftIO D.getTemporaryDirectory >>= resolveDir'
 
-#if MIN_VERSION_directory(1,2,3)
 -- | Obtain the paths to special directories for storing user-specific
 -- application data, configuration, and cache files, conforming to the
 -- <http://standards.freedesktop.org/basedir-spec/basedir-spec-latest.html XDG Base Directory Specification>.
 -- Compared with 'getAppUserDataDir', this function provides a more
 -- fine-grained hierarchy as well as greater flexibility for the user.
 --
--- It also works on Windows, although in that case 'XdgData' and 'XdgConfig'
--- will map to the same directory.
+-- It also works on Windows, although in that case 'D.XdgData' and
+-- 'D.XdgConfig' will map to the same directory.
 --
 -- Note: The directory may not actually exist, in which case you would need
 -- to create it with file mode @700@ (i.e. only accessible by the owner).
@@ -855,14 +882,30 @@ getTempDir = liftIO D.getTemporaryDirectory >>= resolveDir'
 -- @since 1.2.1
 
 getXdgDir :: MonadIO m
-  => XdgDirectory      -- ^ Which special directory
+  => D.XdgDirectory      -- ^ Which special directory
   -> Maybe (Path Rel Dir)
      -- ^ A relative path that is appended to the path; if 'Nothing', the
      -- base path is returned
   -> m (Path Abs Dir)
 getXdgDir xdgDir suffix =
   liftIO $ (D.getXdgDirectory xdgDir $ maybe "" toFilePath suffix) >>= parseAbsDir
-#endif
+
+-- | Similar to 'getXdgDir' but retrieves the entire list of XDG
+-- directories.
+--
+-- On Windows, 'D.XdgDataDirs' and 'D.XdgConfigDirs' usually map to the same
+-- list of directories unless overridden.
+--
+-- Refer to the docs of 'D.XdgDirectoryList' for more details.
+--
+-- @since 1.5.0
+
+getXdgDirList
+  :: MonadIO m
+  => D.XdgDirectoryList         -- ^ Which special directory list
+  -> m [Path Abs Dir]
+getXdgDirList xdgDirList =
+  liftIO (D.getXdgDirectoryList xdgDirList >>= mapM parseAbsDir)
 
 ----------------------------------------------------------------------------
 -- Path transformation
@@ -874,11 +917,11 @@ class AnyPath path where
 
   -- | Type of absolute version of the given @path@.
 
-  type AbsPath path :: *
+  type AbsPath path :: Type
 
   -- | Type of relative version of the given @path@.
 
-  type RelPath path :: *
+  type RelPath path :: Type
 
   -- | Make a path absolute and remove as many indirections from it as
   -- possible. Indirections include the two special directories @.@ and
@@ -1158,16 +1201,129 @@ findFilesWith f (d:ds) file = do
 ----------------------------------------------------------------------------
 -- Symbolic links
 
+-- | Create a /file/ symbolic link. The target path can be either absolute
+-- or relative and need not refer to an existing file. The order of
+-- arguments follows the POSIX convention.
+--
+-- To remove an existing file symbolic link, use 'removeFile'.
+--
+-- Although the distinction between /file/ symbolic links and /directory/
+-- symbolic links does not exist on POSIX systems, on Windows this is an
+-- intrinsic property of every symbolic link and cannot be changed without
+-- recreating the link. A file symbolic link that actually points to a
+-- directory will fail to dereference and vice versa. Moreover, creating
+-- symbolic links on Windows may require privileges unavailable to users
+-- outside the Administrators group. Portable programs that use symbolic
+-- links should take both into consideration.
+--
+-- On Windows, the function is implemented using @CreateSymbolicLink@. Since
+-- 1.3.3.0, the @SYMBOLIC_LINK_FLAG_ALLOW_UNPRIVILEGED_CREATE@ flag is
+-- included if supported by the operating system. On POSIX, the function
+-- uses @symlink@ and is therefore atomic.
+--
+-- Windows-specific errors: This operation may fail with
+-- 'System.IO.Error.permissionErrorType' if the user lacks the privileges to
+-- create symbolic links. It may also fail with
+-- 'System.IO.Error.illegalOperationErrorType' if the file system does not
+-- support symbolic links.
+--
+-- @since 1.5.0
+
+createFileLink
+  :: MonadIO m
+  => Path b0 File               -- ^ Path to the target file
+  -> Path b1 File               -- ^ Path to the link to be created
+  -> m ()
+createFileLink = liftD2 D.createFileLink
+
+-- | Create a /directory/ symbolic link. The target path can be either
+-- absolute or relative and need not refer to an existing directory. The
+-- order of arguments follows the POSIX convention.
+--
+-- To remove an existing directory symbolic link, use 'removeDirLink'.
+--
+-- Although the distinction between /file/ symbolic links and /directory/
+-- symbolic links does not exist on POSIX systems, on Windows this is an
+-- intrinsic property of every symbolic link and cannot be changed without
+-- recreating the link. A file symbolic link that actually points to a
+-- directory will fail to dereference and vice versa. Moreover, creating
+-- symbolic links on Windows may require privileges unavailable to users
+-- outside the Administrators group. Portable programs that use symbolic
+-- links should take both into consideration.
+--
+-- On Windows, the function is implemented using @CreateSymbolicLink@ with
+-- @SYMBOLIC_LINK_FLAG_DIRECTORY@. Since 1.3.3.0, the
+-- @SYMBOLIC_LINK_FLAG_ALLOW_UNPRIVILEGED_CREATE@ flag is also included if
+-- supported by the operating system. On POSIX, this is an alias for
+-- 'createFileLink' and is therefore atomic.
+--
+-- Windows-specific errors: This operation may fail with
+-- 'System.IO.Error.permissionErrorType' if the user lacks the privileges to
+-- create symbolic links. It may also fail with
+-- 'System.IO.Error.illegalOperationErrorType' if the file system does not
+-- support symbolic links.
+--
+-- @since 1.5.0
+
+createDirLink
+  :: MonadIO m
+  => Path b0 Dir                -- ^ Path to the target directory
+  -> Path b1 Dir                -- ^ Path to the link to be created
+  -> m ()
+createDirLink target' dest' = do
+  let target = toFilePath target'
+      dest = F.dropTrailingPathSeparator (toFilePath dest')
+  liftIO $ D.createDirectoryLink target dest
+
+-- | Remove an existing /directory/ symbolic link.
+--
+-- On Windows, this is an alias for 'removeDir'. On POSIX systems, this is
+-- an alias for 'removeFile'.
+--
+-- See also: 'removeFile', which can remove an existing /file/ symbolic link.
+--
+-- @since 1.5.0
+
+removeDirLink
+  :: MonadIO m
+  => Path b Dir                 -- ^ Path to the link to be removed
+  -> m ()
+removeDirLink = liftD D.removeDirectoryLink
+
+-- | Retrieve the target path of either a file or directory symbolic link.
+-- The returned path may not exist, and may not even be a valid path.
+--
+-- On Windows systems, this calls @DeviceIoControl@ with
+-- @FSCTL_GET_REPARSE_POINT@. In addition to symbolic links, the function
+-- also works on junction points. On POSIX systems, this calls @readlink@.
+--
+-- Windows-specific errors: This operation may fail with
+-- 'System.IO.Error.illegalOperationErrorType' if the file system does not
+-- support symbolic links.
+--
+-- @since 1.5.0
+
+getSymlinkTarget
+  :: MonadIO m
+  => Path b t                   -- ^ Symlink path
+  -> m FilePath
+getSymlinkTarget = liftD (D.getSymbolicLinkTarget . F.dropTrailingPathSeparator)
+
+-- | Check whether the path refers to a symbolic link.  An exception is thrown
+-- if the path does not exist or is inaccessible.
+--
+-- On Windows, this checks for @FILE_ATTRIBUTE_REPARSE_POINT@.  In addition to
+-- symbolic links, the function also returns true on junction points.  On
+-- POSIX systems, this checks for @S_IFLNK@.
+--
+-- @since 1.5.0
+
 -- | Check if the given path is a symbolic link.
 --
 -- @since 1.3.0
 
 isSymlink :: MonadIO m => Path b t -> m Bool
-isSymlink p = liftIO (P.isSymbolicLink <$> P.getSymbolicLinkStatus path)
-  where
-    -- NOTE: To be able to correctly check whether it is a symlink or not we
-    -- have to drop the trailing separator from the dir path.
-    path = F.dropTrailingPathSeparator (toFilePath p)
+isSymlink = liftD (D.pathIsSymbolicLink . F.dropTrailingPathSeparator)
 
 ----------------------------------------------------------------------------
 -- Temporary files and directories
@@ -1381,8 +1537,6 @@ copyPermissions = liftD2 D.copyPermissions
 ----------------------------------------------------------------------------
 -- Timestamps
 
-#if MIN_VERSION_directory(1,2,3)
-
 -- | Obtain the time at which the file or directory was last accessed.
 --
 -- The operation may fail with:
@@ -1457,7 +1611,6 @@ setAccessTime = liftD2' D.setAccessTime
 
 setModificationTime :: MonadIO m => Path b t -> UTCTime -> m ()
 setModificationTime = liftD2' D.setModificationTime
-#endif
 
 -- | Obtain the time at which the file or directory was last modified.
 --
